#include <ev.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <poll.h>
#include <ctype.h>
#include <libgen.h>
#include <time.h>
#include <errno.h>
#include <stddef.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/uio.h>
#include <net/if.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <pthread.h>

#include <linux/can.h>
#include <linux/can/raw.h>
#include "lib.h"

#define container_of(ptr, type, member) ({ \
                (type *)( (char *)ptr - offsetof(type,member) );})

#define SECOND (1000000LL)
#define MSECOND (1000LL)

#define MEASURE_TIEMOUT_EN	1

#define MAX_DEVICE	4
#define STOP_ON_ERR	1

#define MAX_PACKET	4	//100
#define FIX_DATA	1

//#define DEBUG(x)      x
#define DEBUG(x)

struct can_pack {
	char device[16];
	ev_io read_io, write_io;
	ev_timer timeout_watcher;

	struct canfd_frame frame_tx[MAX_PACKET], frame_rx[MAX_PACKET];
	unsigned int tx_idx, rx_idx;
	unsigned int current_rate;
	unsigned int timeout;

	unsigned long long tx_count, rx_count;
	unsigned long long tx_count_old, rx_count_old;
	struct timeval start;
};

static struct can_pack m_device[MAX_DEVICE];
static ev_timer show_status_watcher;

static int m_baudrate = 1000000;
static int m_format = 0;	//0: sff, 1: eff
static int m_tx_bytes = 8;
static int m_logger_timer = 30;
static int m_tx_only = 0;

static int en = 1;

static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;

static void time_start(struct timeval *tv)
{
	gettimeofday(tv, NULL);
}

static time_t time_delta(struct timeval *start, struct timeval *stop)
{
	time_t secs, usecs;

	secs = stop->tv_sec - start->tv_sec;
	usecs = stop->tv_usec - start->tv_usec;
	if (usecs < 0) {
		secs--;
		usecs += SECOND;
	}

	return secs * SECOND + usecs;
}

static time_t time_stop(struct timeval *begin)
{
	struct timeval end;
	gettimeofday(&end, NULL);
	return time_delta(begin, &end);
}

static void show_status_timer_cb(EV_P_ ev_timer * w, int revents)
{
	int i;
	unsigned long long elapse;
	int rx_packets_per_sec;
	int tx_packets_per_sec;
	static int count = 0;

	for (i = 0; i < 1; ++i) {
#if 0
		elapse = time_stop(&m_device[i].start);

		tx_packets_per_sec = m_device[i].tx_count * SECOND / elapse;
		rx_packets_per_sec = m_device[i].rx_count * SECOND / elapse;
#else
		pthread_mutex_lock(&mutex);

#if 0
		if (!m_device[i].tx_count_old)
			m_device[i].tx_count_old = m_device[i].tx_count;

		if (m_device[i].rx_count_old)
			m_device[i].rx_count_old = m_device[i].rx_count;
#endif
		rx_packets_per_sec = m_device[i].rx_count - m_device[i].rx_count_old;
		tx_packets_per_sec = m_device[i].tx_count - m_device[i].tx_count_old;

		m_device[i].rx_count_old = m_device[i].rx_count;
		m_device[i].tx_count_old = m_device[i].tx_count;

		pthread_mutex_unlock(&mutex);
#endif

		if (!m_logger_timer) {
			printf("%s: %llu %llu tx:%d, rx:%d, count: %d\n",
			       m_device[i].device,
			       m_device[i].tx_count, m_device[i].rx_count,
			       tx_packets_per_sec, rx_packets_per_sec, count);
		}

		if (m_logger_timer && count > m_logger_timer) {
			printf("%d,%d,%d,%d,%d,%d\n", m_baudrate, m_format, m_tx_bytes,
			       tx_packets_per_sec, rx_packets_per_sec, m_tx_only);

			close(m_device[i].write_io.fd);
			close(m_device[i].read_io.fd);
/*
			printf("%s: %llu %llu tx:%d, rx:%d, count: %d\n",
			       m_device[i].device,
			       m_device[i].tx_count, m_device[i].rx_count,
			       tx_packets_per_sec, rx_packets_per_sec, count);
*/
			exit(0);
		}

		if (!en && count > 10) {
			time_start(&m_device[i].start);
			en = 1;
		}

		count++;
	}
}

static void read_cb(EV_P_ ev_io * w, int revents)
{
	struct can_pack *device = container_of(w, struct can_pack, read_io);
	char buf[CL_CFSZ];	/* max length */
	int j, nbytes;
	struct canfd_frame frame_rx;

	j = device->rx_idx;

	while (1) {
		nbytes = read(w->fd, &frame_rx, CAN_MTU);

		if (nbytes == CAN_MTU)
			break;
		else if (nbytes >= 0)
			printf("%s pending: %d\n", __func__, nbytes);
		else {
			printf("%s: error: %d %d\n", __func__, nbytes, errno);
			exit(0);
		}
	}

	device->rx_idx = (device->rx_idx + 1) % MAX_PACKET;

	DEBUG(sprint_canframe(buf, &frame_rx, 0, CAN_MAX_DLEN));
	DEBUG(printf("rx j: %d, %s\n", j, buf));

	//printf("read rx ok\n");

	/* compare protocol (2.0A/B) */
	if ((device->frame_tx[j].can_id & CAN_EFF_FLAG) != (frame_rx.can_id & CAN_EFF_FLAG)) {
		printf("%s: protocol error\n", device->device);
		exit(0);
	}

	if (device->frame_tx[j].len != frame_rx.len) {
		printf("%s: data len mismatch: %d %d\n", device->device,
		       device->frame_tx[j].len, frame_rx.len);
		exit(0);
	}

	if (device->frame_tx[j].can_id != frame_rx.can_id) {
		printf("%s: id error: %x %x\n", device->device,
		       device->frame_tx[j].can_id, frame_rx.can_id);
		exit(0);
	}
#if 0
	for (i = 0; i < device->frame_tx[j].len; ++i) {
		if (device->frame_tx[j].data[i] != device->frame_rx[j].data[i]) {
			printf("%s: data error: %d\n", device->device, i);

			sprint_canframe(buf, &device->frame_tx[j], 0, CAN_MAX_DLEN);
			printf("TX: %s\n", buf);

			sprint_canframe(buf, &device->frame_rx[j], 0, CAN_MAX_DLEN);
			printf("RX: %s\n", buf);

			exit(0);
		}
	}
#else
	if (memcmp(device->frame_tx[j].data, frame_rx.data, device->frame_tx[j].len)) {
		sprint_canframe(buf, &device->frame_tx[j], 0, CAN_MAX_DLEN);
		printf("TX: %s\n", buf);

		sprint_canframe(buf, &frame_rx, 0, CAN_MAX_DLEN);
		printf("RX: %s\n", buf);
		exit(0);
	}
#endif

	if (en) {
		pthread_mutex_lock(&mutex);
		device->rx_count++;
		pthread_mutex_unlock(&mutex);
	}

	ev_io_stop(EV_A_ w);
	ev_io_start(EV_A_ & device->write_io);
}

static void write_cb(EV_P_ ev_io * w, int revents)
{
	struct can_pack *device = container_of(w, struct can_pack, write_io);
	char buf[CL_CFSZ];	/* max length */
	int j, nbytes;

	j = device->tx_idx;

	while (1) {
		DEBUG(sprint_canframe(buf, &device->frame_tx[j], 0, CAN_MAX_DLEN));
		DEBUG(printf("tx j: %d, %s\n", j, buf));

		nbytes = write(w->fd, &device->frame_tx[j], CAN_MTU);
		if (nbytes == CAN_MTU)
			break;
		else if (nbytes >= 0)
			printf("%s pending: %d\n", __func__, nbytes);
		else {
			printf("%s: error: %d %d\n", __func__, nbytes, errno);
			exit(0);
		}
	}

	device->tx_idx = (device->tx_idx + 1) % MAX_PACKET;

	//printf("write tx ok\n");

	/* log CAN frame with absolute timestamp & device */
	//sprint_canframe(buf, &device->frame_tx[j], 0, CAN_MAX_DLEN);
	//printf("TX: %s\n", buf);

	if (en) {
		pthread_mutex_lock(&mutex);
		device->tx_count++;
		pthread_mutex_unlock(&mutex);
	}

	if (!m_tx_only) {
		ev_io_stop(EV_A_ w);
		ev_io_start(EV_A_ & device->read_io);
	}
}

static void timeout_cb(EV_P_ ev_timer * w, int revents)
{
}

char *m_init_string[] = {
	"ip link set can%d down > /dev/null 2>&1\n",
	"ip link set can%d type can restart-ms 0 > /dev/null 2>&1\n",
	"ip link set can%d type can bitrate %d sample-point 0.%d > /dev/null 2>&1\n",
	"ip link set can%d txqueuelen 1000 > /dev/null 2>&1\n",
	"ip link set can%d type can loopback off > /dev/null 2>&1\n",
	"ip link set can%d type can one-shot off > /dev/null 2>&1\n",
	"ip link set can%d type can listen-only off > /dev/null 2>&1\n",
	"ip link set can%d type can presume-ack off > /dev/null 2>&1\n",
	"ip link set can%d type can berr-reporting on > /dev/null 2>&1\n",
	"tc qdisc add dev can%d root handle 1: pfifo > /dev/null 2>&1",
	"ip link set can%d up\n",
	//"sleep 1\n",
};

int m_baudrate_table[] = { 1000000,
	800000,
	500000,
	250000,
	125000,
	100000,
	50000,
	20000,
	10000,
};

int m_percent_table[] = {
	75,
	80,
	875,
	875,
	875,
	875,
	875,
	875,
	875,
};

static int init_canbus(const char *device, int dataset, int baudrate)
{

	struct sockaddr_can addr;
	struct ifreq ifr;
	int s[2] = { -1, -1 };
	int loopback = 0;
	int i, j;
	int idx;
	char buf[1024];
	//int can_idx[] = { 2, 0 };

	for (idx = 0; idx < sizeof(m_baudrate_table) / sizeof(int); ++idx) {
		if (m_baudrate_table[idx] == baudrate)
			break;
	}

	if (idx == sizeof(m_baudrate_table)) {
		printf("baudrate error: %d\n", baudrate);
		exit(0);
	}
	//printf("idx: %d\n", idx);
	//printf("baudrate: %d\n", m_baudrate_table[idx]);
	//printf("percentage: %d\n", m_percent_table[idx]);

#if 1
	for (j = 0; j < 2; ++j) {
		for (i = 0; i < sizeof(m_init_string) / sizeof(char *); ++i) {
			sprintf(buf, m_init_string[i], j, m_baudrate_table[idx],
				m_percent_table[idx]);
			system(buf);
			//printf(buf);
		}
	}
#endif
	//exit(0);

	//sprintf(m_device[dataset].device, "%s%d", device, dataset);
	//printf("device: %s\n", m_device[dataset].device);

	for (i = 0; i < 2; ++i) {
		//sprintf(m_device[dataset].device, "%s%d", device, can_idx[i]);
		sprintf(m_device[dataset].device, "%s%d", device, i);

		if ((s[i] = socket(PF_CAN, SOCK_RAW, CAN_RAW)) < 0) {
			perror("socket");
			exit(1);
		}

		addr.can_family = AF_CAN;

		strcpy(ifr.ifr_name, m_device[dataset].device);
		if (ioctl(s[i], SIOCGIFINDEX, &ifr) < 0) {
			perror("SIOCGIFINDEX");
			exit(1);
		}

		addr.can_ifindex = ifr.ifr_ifindex;

		//setsockopt(s, SOL_CAN_RAW, CAN_RAW_FILTER, NULL, 0);
		setsockopt(s[i], SOL_CAN_RAW, CAN_RAW_LOOPBACK, &loopback, sizeof(loopback));

		if (bind(s[i], (struct sockaddr *)&addr, sizeof(addr)) < 0) {
			perror("bind");
			break;
		}
	}

	if (s[0] < 0 || s[1] < 0) {
		if (s[0] >= 0)
			close(s[0]);
		if (s[1] >= 0)
			close(s[1]);

		return -1;
	}

	if (!m_logger_timer) {
		//printf("tx device: can%d\n", can_idx[1]);
		//printf("rx device: can%d\n", can_idx[0]);

		printf("tx device: can%d\n", 1);
		printf("rx device: can%d\n", 0);
	}

	ev_io_init(&m_device[dataset].write_io, write_cb, s[1], EV_WRITE);
	ev_io_init(&m_device[dataset].read_io, read_cb, s[0], EV_READ);

	ev_timer_init(&m_device[dataset].timeout_watcher, timeout_cb, 1, 1);

	for (i = 0; i < MAX_PACKET; ++i) {
		m_device[dataset].frame_tx[i].can_id = rand();

		if (!m_format) {
			m_device[dataset].frame_tx[i].can_id &= CAN_SFF_MASK;
		} else {
			m_device[dataset].frame_tx[i].can_id &= CAN_EFF_MASK;
			m_device[dataset].frame_tx[i].can_id |= CAN_EFF_FLAG;
		}

#if FIX_DATA
		m_device[dataset].frame_tx[i].len = m_tx_bytes;
#else
		m_device[dataset].frame_tx[i].len = rand() % 9;
#endif

		if (m_device[dataset].frame_tx[i].len == 0) {
			m_device[dataset].frame_tx[i].can_id |= CAN_RTR_FLAG;
		} else {
			for (j = 0; j < m_device[dataset].frame_tx[i].len; ++j)
				m_device[dataset].frame_tx[i].data[j] = rand() & 0xff;
		}
	}

	return 0;
}

void usage(char *argv[])
{
	printf("%s <baud> <format> <tx byte> <timeout> <tx only>\n", argv[0]);
	printf("	<format>: 0: sff, 1: eff\n");
}

int main(int argc, char *argv[])
{
	struct ev_loop *loop = EV_DEFAULT;
	char tty_path[] = "can";
	int i;

	ev_timer_init(&show_status_watcher, show_status_timer_cb, 2, 1);

	srand(time(NULL));

	if (argc < 6) {
		usage(argv);
		return 0;
	}

	m_baudrate = strtol(argv[1], NULL, 10);
	m_format = strtol(argv[2], NULL, 10);
	m_tx_bytes = strtol(argv[3], NULL, 10);
	m_logger_timer = strtol(argv[4], NULL, 10);

	for (i = 0; i < sizeof(m_baudrate_table) / sizeof(m_baudrate_table[0]); ++i) {
		if (m_baudrate == m_baudrate_table[i])
			break;
	}

	if (i == sizeof(m_baudrate_table) / sizeof(m_baudrate_table[0])) {
		printf("baudrate error: %d\n", m_baudrate);
		exit(0);
	}

	if (m_format != 0 && m_format != 1) {
		printf("m_format error: %d\n", m_format);
		exit(0);
	}

	if (m_tx_bytes < 0 || m_tx_bytes > 8) {
		printf("m_tx_bytes error: %d\n", m_tx_bytes);
		exit(0);
	}

	if (m_logger_timer < 0) {
		printf("m_logger_timer error: %d\n", m_logger_timer);
		exit(0);
	}

	m_tx_only = strtol(argv[5], NULL, 10);
	if (m_tx_only != 0 && m_tx_only != 1) {
		printf("m_tx_only error: %d\n", m_tx_only);
		exit(0);
	}

	for (i = 0; i < 1; ++i) {
		init_canbus(tty_path, i, m_baudrate);
	}

	ev_timer_start(EV_A_ & show_status_watcher);

	//test();
	for (i = 0; i < 1; ++i) {
		ev_io_stop(EV_A_ & m_device[i].write_io);
		ev_io_stop(EV_A_ & m_device[i].read_io);

		time_start(&m_device[i].start);
		ev_io_start(EV_A_ & m_device[i].write_io);
	}

	ev_run(EV_A_ 0);

	return 0;
}
